#include "global.hpp"

namespace cr {


  // returns true iff b can be split at it's head
  bool is_splittable(const edge_p& b){
    const vertex_p& v(b->partner);
    // if v is a leaf, then it's not splittable
    if(v->degree() <= 1) return false;
    // if v has nldeg > 2, then v is not splittable
    if(v->nldeg() > 2) return false;
    // if any of v's neighbors has nldeg > 2, then v is not splittable
    for(list<edge>::const_iterator w = v->adj_list.begin(); w != v->adj_list.end(); ++w)
      if(w->partner->nldeg() > 2) return false;
    return true;
  }
  
  // apply the split rule to a specific bridge b, which is splittable
  // NOTE: this might invalidate (delete) some bridges
  void apply_split_rule(instance& I, edge_p b){
    const vertex_p& v(b->partner);
    const vertex_p& u(b->get_reversed()->partner);
    // delete the bridge
    I.g.delete_edge(b);
    // and add a new deg-1 neighbor to u
    const vertex_p vprime(I.g.add_vertex_fast((string)*v + "'"));
    I.g.add_edge_fast(vprime, u);
    // apply TRRs from vprime upwards
  }
  
  // split rule: find bridge {u,v} such that deg(v)>1 and all w in N[v] have nldeg < 3, then split {u,v} from v
  bool apply_split_rule(instance& I){
    solution_t sol;
    bool change = false;

    // first, get all bridges
    edgelist bridges = I.g.get_bridges();
    // for each bridge, see if the rule applies
    for(edgelist::iterator b = bridges.begin(); b != bridges.end();)
      if(is_splittable(*b)){
        // remember to advance b before deleting anything
        apply_split_rule(I, *(b++));
        change = true;
      } else if(is_splittable((*b)->get_reversed())){
        // remember to advance b before deleting anything
        apply_split_rule(I, (*(b++))->get_reversed());
        change = true;
      } else ++b;
    // finally, apply the TRRs if something happened
    return change;
  }




// ==== Reduction Rule 2.1 ======

  // verify that there are no cyc-deg-3 vertices reachable from v using only non-bridges
  bool is_leaf_in_comp_tree(const vertex_p& v, const vertex_p& camefrom, const size_t dfs_id){
    // if we've been called for this component before but didn't say YAY, don't say YAY now...
    if(camefrom == v){ // if this is the first vertex in this component
      if(v->dfs_id == dfs_id) return false;
    } else {
      if(v->dfs_id == dfs_id) return true;
      // if v is not the first vertex, but incident to B bridges, then return failure
      if(v->is_incident_to_Bbridge()) return false;
    }
    DEBUG2(cout << "exploring "<<*v<< endl);
    v->dfs_id = dfs_id;

    edgelist nh(v->get_cyclic_neighbors());

    bool akku = true;
    for(auto i = nh.begin(); i != nh.end(); ++i)
      if(((*i)->partner != camefrom) && ((*i)->partner->dfs_id != dfs_id))
        akku &= is_leaf_in_comp_tree((*i)->partner, v, dfs_id);
    return akku;
  }
  // return the only cyc-core-deg3 vertex in a leaf of the component tree
  vertex_p get_comptree_leaf(graph& g){
    for(vertex_p v = g.vertices.begin(); v != g.vertices.end(); ++v)
      if((v->non_bridge_degree() > 1) && (v->is_incident_to_Bbridge())){
        DEBUG2(cout << "new vertex to check out: "<<*v<<endl);
        size_t dfs_id = g.get_dfs_id();
        if(is_leaf_in_comp_tree(v, v, dfs_id)){
          DEBUG2(cout << "found that component of "<<*v<<" is a comptree leaf"<<endl);
          return v;
        }
      }
    DEBUG2(cout << "couldn't find a comptree leaf, returning first vertex on the cyclic core"<<endl);
    // if there are no cyc-core-deg3 vertices, just return any vertex on the cyclic core
    for(vertex_p v = g.vertices.begin(); v != g.vertices.end(); ++v)
      if(v->is_on_cyclic_core()) return v;

    // if this is reached, then we have a forest, but this should have been deleted by reduction
    FAIL("epic fail in get_comptree_leaf: forgot to apply tree reductions?");
  }
  edge_p get_farthest_edge_on_cycle(const vertex_p& u){
    edgelist nh(u->get_cyclic_neighbors());
    if(nh.size() > 2) FAIL("epic fail in get_farthest_edge_on_cycle");
    edge_p left = nh.front();
    edge_p right = nh.back();
    do{
      left = get_next_on_cycle(left);
      right = get_next_on_cycle(right);
    } while((left != right->get_reversed()) && (left->partner != right->partner));
    return left;
  }
  inline bool is_dirty(const vertex_p& v){
    return !(v->trr_infos.ptwos.empty());
  }

  // apply the final reduction rule when nothing else is applicable anymore
  solution_t apply_final_RR(instance& I){
    // make sure bridges are marked correctly
    I.g.mark_bridges();
    // get a leaf of the component tree of I.g
    vertex_p u = get_comptree_leaf(I.g);
    DEBUG2(cout << "got component-tree leaf "<<*u<<endl);

    // get all neighbors of u along non-bridges
    edgelist nh(u->get_cyclic_neighbors());
    if(nh.size() != 2) FAIL("epic fail in final_RR");

    // the two cyclic neighbors of u are v and w
    vertex_p w(nh.back()->partner);
    vertex_p v(nh.front()->partner);

    // make sure that v is dirty whenever w is dirty
    if(is_dirty(w) && !is_dirty(v)) swap(v, w);

    // apply the actual reduction:
    solution_t sol;
    if((v->non_bridge_degree() == 2) && is_dirty(w)){
      // if v cyclic degree of v is exactly 2 and w is dirty, then delete farthest edge from u
      edge_p a = get_farthest_edge_on_cycle(u);
      DEBUG3(cout << "RR1: edge deletion: "<<*a<<endl);
      I.delete_edge(a, sol);
    } else {
      // otherwise, delete {u,v}
      edge_p a(v->first_non_bridge_neighbor_except(w));
      DEBUG3(cout << "RR1: edge deletion: "<<*a<<endl);
      I.delete_edge(a, sol);
    }
    return sol;
  }


  // NOTE: Tarjan's algorithm already computes

  // apply the B-bridge rule:
  // Let T' be a proper subtree of the component tree T of G whose unique incident B-bridge in G is {u,v} (with v in T').
  // Let S1 , S2' , and S3 denote optimal solutions of G[T'], G[T'-v], and G[T'+u] and
  // let S2 := S2 cup {{v, w} in E(G) | w = u}.
  // (a) If |S1 | < |S3 |, then apply S1 + {u, v} to G.
  // (b) If not (a), but |S2 | = |S3 |, then apply S2 to G.
  // (c) If neither (a) nor (b), then apply S3 to G.
  solution_t apply_Bbridge_rule(instance& I){
    // first, get me a good B-bridge: the score of a B-bridge is how evenly it splits the graph G
    edge_p best_bridge;
    uint highscore = 0;
    // get the list of Bbridges, NOTE: this marks all bridges, we'll play around with bridges_marked a bit since we know what we're doing
    edgelist B(I.g.get_Bbridges());
    for(edgelist::iterator e = B.begin(); e != B.end(); ++e){
      // get a copy of the B bridge in a representation that allows deleting it
      const vertex_p& u((*e)->get_reversed()->partner);
      const vertex_p& v((*e)->partner);
      // delete the B bridge
      I.g.delete_edge(*e);
      // this should split the graph in EXACTLY 2 connected components, so count the size of one of them
      uint comp1 = I.g.component_size(I.g.vertices.begin());
      uint comp2 = I.g.vertices.size() - comp1;
      // swap the sizes if necessary to have comp1 <= comp2
      if(comp1 > comp2) {uint temp = comp1; comp1 = comp2; comp2 = temp;}
      
      // finally, reinsert the B bridge and act as if nothing happened
      edge_p b(I.g.add_edge_fast(u, v));
      b->mark_bridge();
      I.g.bridges_marked = true;

      // remember to save the score and the bridge
      if(comp2 - comp1 > highscore){
        highscore = comp2 - comp1;
        best_bridge = b;
      }
    }
//#error CONTINUE HERE
  }

}
