#include "prr.hpp"
#include "trr.hpp"
#include <sstream>

namespace cr{

  edge_p skip_ahead_non_intrusive(uint dfs_id, edge_p& direction, uint *length){
    DEBUG2(cout << "skipping non-intrusively in direction "<<*direction<<endl);
    const vertex_p& to(direction->partner);
    
    // if we looped, then return
    if(to->dfs_id == dfs_id) return direction;
    // if a cyc-deg-3 vertex is reached, return it
    if(to->cyc_core_degree() > 2) return direction;

    // otherwise, set the dfs_id and skip ahead in direction
    to->dfs_id = dfs_id;
    
    // if there are no more cyclic neighbors, return the current vertex
    const edge_p& next(get_next_on_deg2path(direction));
    if(next == to->adj_list.end()) return direction;
    direction = next;

    // advance the length-counter
    if(length) ++(*length);

    return skip_ahead_non_intrusive(dfs_id, direction, length);

  }

  // add a leaf to v, keeping subtree_NH up to date; returns edge to the new leaf
  edge_p& add_leaf(graph& g, const vertex_p& v, const string name, const uint dfs_id){
    // create the vertex
    const vertex_p w(g.add_vertex_fast(name));
    w->dfs_id = dfs_id;
    // connect it to v
    const edge_p wv(g.add_edge_fast(w, v));

    // update w
    w->set_parent(wv);

    // update v
    v->subtree_NH++;

    // and return edge to the new leaf
    return wv->get_reversed();
  }

  // copy v's complete pendant tree to vprime (including dfs_id's)
  void copy_pendant(graph& g, const vertex_p& v, const vertex_p& vprime){
    // step 1: copy leaves
    for(edge_pp e = v->trr_infos.leaves.begin(); e != v->trr_infos.leaves.end(); ++e)
      vprime->trr_infos.leaves.push_back(add_leaf(g, vprime, (string)*((*e)->partner) + "'", (*e)->partner->dfs_id));
    
    // step 2: copy P2s
    for(edge_pp e = v->trr_infos.ptwos.begin(); e != v->trr_infos.ptwos.end(); ++e){
      const edge_p to_leaf(add_leaf(g, vprime, (string)*((*e)->partner) + "'", (*e)->partner->dfs_id));
      vprime->trr_infos.ptwos.push_back(to_leaf);
      copy_pendant(g, (*e)->partner, to_leaf->partner);
    }

    // step 3: copy Ys
    for(edge_pp e = v->trr_infos.ygraphs.begin(); e != v->trr_infos.ygraphs.end(); ++e){
      const edge_p to_leaf(add_leaf(g, vprime, (string)*((*e)->partner) + "'", (*e)->partner->dfs_id));
      vprime->trr_infos.ygraphs.push_back(to_leaf);
      copy_pendant(g, (*e)->partner, to_leaf->partner);
    }
  }




// ***********************************
// * NEW functions for applying PRRs *
// **********************************f

  bool prr1_applicable(const vertex_p& v){
    // v is v_i;  consider each of v's cyclic neighbors as v_{i+1}
    const edgelist nh(v->get_cyclic_neighbors());
    for(auto e = nh.begin(); e != nh.end(); ++e){
      const vertex_p& w((*e)->partner);
      // w should be an inner vertex
      if(w->cyc_core_degree() != 2) continue;
      // 1.: T^w is neither sigleton nor Y
      if(!w->pendant_is_single() && !w->pendant_is_Y()) return true;
      // 2.: w is degree-2 and the next vertex u after w i an inner vertex with singleton or Y as pendant
      if(w->pendant_is_single()){
        const vertex_p& u(w->first_cyclic_core_neighbor_except(v)->partner);
        if(u->cyc_core_degree() == 2) // lazy evaluation might save some CPU-cycles...
           if(u->pendant_is_single() || u->pendant_is_Y()) return true;
      }
    }
    return false;
  }

  bool prr2_applicable(const vertex_p& v){
     // v is v_i;  consider each of v's cyclic neighbors as v_{i+1}
    const edgelist nh(v->get_cyclic_neighbors());
    for(auto e = nh.begin(); e != nh.end(); ++e){
      const vertex_p& w((*e)->partner);
      // w should be an inner vertex
      if(w->cyc_core_degree() > 2) continue;
      // 1. T^w is Y
      if(w->pendant_is_Y()) return true;
      // 2. the vertex before v and w both have degree-2
      if(w->pendant_is_single()){
        const vertex_p& u(v->first_cyclic_core_neighbor_except(w)->partner);
        if((u->cyc_core_degree() == 2) && u->pendant_is_single()) return true;
        
        // 3. the vertex x after w is inner and its pendant is neither single nor Y
        const vertex_p& x(w->first_cyclic_core_neighbor_except(v)->partner);
        if((x->cyc_core_degree() == 2))
          if(!x->pendant_is_single() && !x->pendant_is_Y()) return true;
      }
    }
    return false;
  }

  bool prr3_applicable(const path_info_t& info){
    // no use applying this to paths of 4 or less inner edges
    // if there are no tokens, pp3 should be applicable
    return ((info.length > 4) && info.generators.empty());
  }

  bool prr4_applicable(const path_info_t& info){
    // prr4 is applicable if there is a token separator and a generator on the path
    return (info.has_separator && !info.generators.empty());
  }

  bool prr5_applicable(const path_info_t& info){
    // prr5 is applicable if there are more than 2 generators and no separator
    return (!info.has_separator && info.generators.size() > 2);
  }

  // do the edge deletion and handle the path_infos
  solution_t perform_prr1(instance& I, const vertex_pp& v, path_info_t& info){
    DEBUG2(cout << "this is PRR1 for "<<*v<<endl);
    solution_t sol;
    
    const edge_p e((*v)->trr_infos.ygraphs.front());
    const vertex_p w(e->partner);

    // remove the Ygraph from v's trr_infos
    (*v)->trr_infos.ygraphs.pop_front();
    (*v)->subtree_NH--;

    // kill the edge to the center of v's Y-graph
    I.delete_edge(e, sol);

    // and delete the split off caterpiller
    I.g.delete_component(w);

    // finally, unregister our Y-graph with the path_info
    info.pendantYs.erase(v);
    
    return sol;
  }

  solution_t perform_prr2(instance& I, const vertex_pp& v, path_info_t& info){
    DEBUG2(cout << "this is PRR2 for "<<*v<<endl);
    solution_t sol;

    // get the two neighbors of v
    const edgelist nh((*v)->get_cyclic_core_neighbors());
    vertex_p u(nh.front()->partner);
    vertex_p w(nh.back()->partner);
    
    DEBUG2(cout << "got cyclic neighbors "<<u<<" and "<<w<<endl);

    // if the path is actually a triangle,
    if(find_edge(u, w) != u->adj_list.end()){
      // then delete both edges incident to v
      I.delete_edges(nh, sol);
      // and mark info invalid
      info.valid = false;
      
      // let the TRRs gnaw on the newfound edge
      // the gnawing should start at the lower one (where degree() - subtree_NH == 1)
      if(u->degree() - u->subtree_NH == 1)
        sol += apply_trrs_upwards_after_cut(I, u, w);
      else
        sol += apply_trrs_upwards_after_cut(I, w, u);
    } else {
      // else skip over v by {u,w} and update u and w (could have become separators)
      sol += (string)**v+"->?";
      I.k--;

      // delete the two incident edges and
      I.g.delete_edges(nh);
      
      // add the skip edge
      I.g.add_edge_fast(u,w);

      // obstusify u & w if they are on the def-2 path, so verification is possible
      // because if (3) matched, then the solution actually has to be transformed,
      // although the size is the same
      if(u->cyc_core_degree() == 2) u->name += "*";
      if(w->cyc_core_degree() == 2) w->name += "*";

      // set the separator if any of the two has become one
      if(!info.has_separator){
        if(u->is_separator()){
          info.separator = u;
          info.has_separator = true;
        } else if(w->is_separator()){
          info.separator = w;
          info.has_separator = true;
        }
      }

      // also, the length of the path just decreased
      info.length--;
    }
    // don't forget to delete the component of v
    I.g.delete_component(*v);
    return sol;
  }

  void perform_prr3(instance& I, path_info_t& info){
    DEBUG2(cout << "this is PRR3 for the path starting at "<<*info.start<< " & ending at " << *info.end << endl);
    // get second to first and second to last vertex (should be different by length > 4)
    const edge_p del_start(get_next_on_deg2path(info.start));
    const vertex_p scnd_first(del_start->partner);
    const vertex_p scnd_last(info.end->get_reversed()->partner);

    // delete everything between scnd_first and scnd_last
    edge_p e1(get_next_on_deg2path(del_start));
    edge_p e2(get_next_on_deg2path(info.end->get_reversed()));
    const vertex_p to_die(e1->partner);
    I.g.delete_edge(e1);
    I.g.delete_edge(e2);
    I.g.delete_component(to_die);

    // then, connect scnd_first and scnd_last
    I.g.add_edge_fast(scnd_first, scnd_last);

    // after this, the path cannot be used for anything anymore
    info.valid = false;
  }

  solution_t perform_prr4(instance& I, path_info_t& info){
    DEBUG2(cout << "this is PRR4 for "<<*info.separator << endl);
    solution_t sol;

    const vertex_p do_not_cross(info.start->get_reversed()->partner);
    vertex_p v(info.separator);
    const edge_p& e(v->first_cyclic_core_neighbor());
    const vertex_p w(e->partner);

    I.g.delete_edge(e);
    // create a copy of v
    vertex_p vprime(I.g.add_vertex_fast((string)*v + "'"));
    // and hang it onto w
    I.g.add_edge_fast(vprime, w);
    // then copy v's pendant to vprime
    copy_pendant(I.g, v, vprime);
    // and, finally, gnaw the 2path away
    DEBUG2(cout << "first, gnaw away from "<< v << " to " << do_not_cross << endl);
    sol += apply_trrs_upwards_after_cut(I, v, do_not_cross);
    DEBUG2(cout << "second, gnaw away from "<< vprime << " to " << do_not_cross<< endl);
    sol += apply_trrs_upwards_after_cut(I, vprime, do_not_cross);

    // oh, right, don't forget to invalidate the path info
    info.valid = false;
    return sol;
  }

  solution_t perform_prr5(instance& I, path_info_t& info){
    DEBUG2(cout << "this is PRR5 for the path starting at "<<*info.start<< " & ending at " << *info.end << endl);
    solution_t sol;
    // if the number of generators is odd, then start erasing BEFORE the first generator, otherwise, BEHIND it
    const edge_p first_to_del(((info.generators.size() & 1) == 1) ?
          info.generators.front()  :   get_next_on_deg2path(info.generators.front())  );
    // the last edge to delete is always the one pointing to the last generator
    const edge_p last_to_del(info.generators.back());
    
    // decrease k and register edges in solution
    // if generators is even, decrease k by |generators|/2 - 1, otherwise by |generators|/2 - 1/2
    // which is the same as always decreasing by floor( |generators|/2 - 1/2 )
    const uint deletes((info.generators.size() - 1) >> 1);
    I.k -= deletes;
    // we don't yet know which edges are going to be deleted..
    for(uint i = 0; i < deletes; ++i)
      sol += "[some edge between " + (string)(*info.generators.front()->partner) + " and " + (string)(*info.generators.back()->partner) + "]";

    // do the graph modification:
    // 0. save the first and last vertex in order to be able to reconnect
    const vertex_p first_vertex(first_to_del->get_reversed()->partner);
    vertex_p last_vertex(last_to_del->partner);
    // and save one vertex to be able to delete the component in between
    const vertex_p comp_to_del(first_to_del->partner);
    // (save the name of first_to_del in case we cut a cycle)
    const string first_to_del_name(*first_to_del);

    // 1. delete the first and last edge,
    I.g.delete_edge(first_to_del);
    I.g.delete_edge(last_to_del);
    // 2. delete the connected component and
    I.g.delete_component(comp_to_del);
    // 3. add a new edge, unless it is already there, in which case we broke a cycle
    edge_p e(find_edge(first_vertex, last_vertex));
    if(e != first_vertex->adj_list.end()){
      // if we broke a cycle, then delete the edge (first_vertex, last_vertex) (destroy the 2-cycle)
      sol += first_to_del_name;
      I.k--;
      // update the TRRs just one edge, and let the parent vertex update the rest
      sol += apply_trrs_upwards_after_cut(I, last_vertex, first_vertex);
    } else I.g.add_edge_fast(first_vertex, last_vertex);
    // mark the info invalid, since no other rule appies to this path any more
    info.valid = false;

    return sol;
  }

// ********************************************
// * NEW overhead functions for applying PRRs *
// ********************************************

#define BUDGET_EXCEEDED {I.k = -1; info.valid = false; return solution_t();}
  solution_t prr12_from_infos(instance& I, path_info_t& info){
    solution_t sol;
    DEBUG2(cout << "checking prr12"<<endl);

    // find a vertex with a Y-pendant and apply PRR1&2
    while(!info.pendantYs.empty() && info.valid){
      const vertex_pp& v(info.pendantYs.begin());

      // in both PRR1 and PRR2, we need at least one edge deletion, so we'll return failure if the budget ran out
      if(prr1_applicable(*v)){
        if(I.k <= 0) BUDGET_EXCEEDED else sol += perform_prr1(I, v, info);
      } else if(prr2_applicable(*v)){
        if(I.k <= 0) BUDGET_EXCEEDED else sol += perform_prr2(I, v, info);
        // if neither prr1 nor prr2 applies, we face a short path with just a Y-graph -> nothing to do
      } else info.valid = false; 
    }
    return sol;
  }
  
  // if there is no token generator on the degree-2-path, then replace the path by a simple P4 (given its not already shorter)
  void prr3_from_infos(instance& I, path_info_t& info){
    DEBUG2(cout << "checking prr3"<<endl);
    if(prr3_applicable(info))
      perform_prr3(I, info);
  }

  // if there is a token separator v on a path with tokens, split the path open at v
  solution_t prr4_from_infos(instance& I, path_info_t& info){
    DEBUG2(cout << "checking prr4"<<endl);
    if(prr4_applicable(info))
      if(I.k <= 0) BUDGET_EXCEEDED else return perform_prr4(I, info);
    else return solution_t();
  }

  // if we have a chained degree-2 path of even (odd) length, compress it to 2 (to 3) generators
  solution_t prr5_from_infos(instance& I, path_info_t& info){
    DEBUG2(cout << "checking prr5"<<endl);
    if(prr5_applicable(info))
      if(I.k <= 0) BUDGET_EXCEEDED else return perform_prr5(I, info);
    else return solution_t();
  }


  // get infos along the deg2-path initiated by e, assumes that e is on a cyc-core-deg2-path
  path_info_t get_path_infos(const edge_p& e, const uint dfs_id){
    path_info_t infos;
    edge_p next(e);


    // set startpoint of path
    infos.start = e;
    // while the next vertex is valid and on the deg2-path
    while((next->partner->cyc_core_degree() == 2) && (next->partner->dfs_id != dfs_id)){
      const vertex_p& v(next->partner);

      DEBUG2(cout << "exploring path along " << *next << " with trr("<<*v<<") = "<< v->trr_infos << endl);
      DEBUG2(cout << "deg("<<*v<<")="<<v->degree()<<", subtree("<<*v<<")="<<v->subtree_NH<< " dfs = " << v->dfs_id << " (vs. "<<dfs_id << ")" << endl);

      // mark this vertex seen
      v->dfs_id = dfs_id;
      
      // count the length
      ++infos.length;

      // analyze the next vertex
      if(!infos.has_separator && v->is_separator()) {
        infos.separator = v;
        infos.has_separator = true;
      }
      if(v->is_generator()) infos.generators.push_back(next);
      // prr6: if v is both generator and separator (has 2 P2), then we just skip the rest of the path!
      // CAUTION: if we break here, then a path that returns may be detected twice and if this path is destroyed, considering the second one segfaults
//      if(v->trr_infos.ptwos.size() > 1) break;
      
      // mark vertices with Y-graph pendants
      if(!v->trr_infos.ygraphs.empty()) infos.pendantYs.push_back(v);

      // advance along the path
      DEBUG2(cout << "next, going to "<< v->first_cyclic_core_neighbor_except(next->get_reversed()->partner) << endl);
      next = get_next_on_deg2path(next);
    }
    DEBUG2(cout << "finished exploring at "<< *next << endl);
    // set the end of the path and mark it valid
    infos.end = next;
    infos.valid = true;

    return infos;
  }

  // act on path information
  // CAUTION: this can destroy alot - use safe iterators!
  solution_t act_on_path_info(instance& I, path_info_t& info){
    // first look for Y-graphs and apply PRR1 & PRR2
    solution_t sol;

    DEBUG2(cout << "acting on " << info << endl);

    sol += prr12_from_infos(I, info);
    if(!info.valid) return sol;

    prr3_from_infos(I, info);
    if(!info.valid) return sol;

    sol += prr4_from_infos(I, info);
    if(!info.valid) return sol;

    sol += prr5_from_infos(I, info);
    return sol;
  }

  // find the first path of v
  edge_p find_first_path(const vertex_p v, const uint dfs_id){
    for(edge_p e = v->adj_list.begin(); e != v->adj_list.end(); ++e)
      if(e->partner->is_on_cyclic_core() && (e->partner->dfs_id != dfs_id))
        return e;
    return v->adj_list.end();
  }
  // find the next path, ignoring info.end and any paths we have seen or are not cyc-core-deg2-paths
  edge_p find_next_path(const path_info_t info, const uint dfs_id){
    // get the current edge and vertex
    edge_p result(info.start);
    const vertex_p& v(result->get_reversed()->partner);
    // advance result until it's either at the end of v's neighbor list or it's valid
    while(true){
      ++result;
      if(result == v->adj_list.end()) break;
      if(result->partner->dfs_id == dfs_id) continue;
      if(!result->partner->is_on_cyclic_core()) continue;
      if(result == info.end->get_reversed()) continue;
    }
    DEBUG2(if(result != v->adj_list.end()) cout << "next is going to be "<< result << endl; else cout << "this is the end"<<endl);
    return result;
  }

  // apply the path reduction rules to the instance I
  solution_t apply_prrs(instance& I){
    solution_t sol;

    // make sure the cyclic core is up to date
    I.g.update_subtree_NH();

    // if there are no cyc-deg3 vertices, we might want to allow starting at a cyc-deg2 vertex
    bool allow_start_at_deg2 = false;

    // exhaustively apply PRRs
    uint old_size;
    do{
      old_size = sol.size();

      // get a new dfs_id from I.g
      uint dfs_id(I.g.get_dfs_id());

      // save the last vertex with cyc_core_degree == 2, in case we have only a cycle
      vertex_p last_cyc;
      bool has_cyc_deg3 = false;
      bool has_cyc_deg2 = false;

      vertex_p v(I.g.vertices.begin());
      // while we haven't seen all vertices
      while(v != I.g.vertices.end()){
        DEBUG2(cout << "considering "<< (string)*v <<endl);

        // make sure this vertex has not been considered yet
        if(v->dfs_id != dfs_id){
          DEBUG2(cout << (string)*v << " is unconsidered with trr: " << v->trr_infos << " & subs: "<< v->subtree_NH << endl);
          // also make sure that it's on a cycle
          if((v->cyc_core_degree() > 2) || (allow_start_at_deg2 && v->cyc_core_degree() == 2)){
            DEBUG2(cout << (string)*v << " is cyclic-core-degree >2"<< endl);
            has_cyc_deg3 = true;

            // mark v as considered now
            v->dfs_id = dfs_id;

            // for each path, get its info act on it
            for(edge_p path_via = find_first_path(v, dfs_id); path_via != v->adj_list.end();){
              // collect the path infos
              path_info_t path_info(get_path_infos(path_via, dfs_id));
              DEBUG2(cout << "found path:"<< path_info << endl);
              // increase the iterator
              path_via = find_next_path(path_info, dfs_id);
              // finally, act on the path
              sol += act_on_path_info(I, path_info);
              DEBUG2(cout << "done acting on this path, solution now " << sol << ", remaining budget: " << I.k << endl);
            }
            DEBUG2(cout << "all paths of "<< *v << " are done, now applying TRRs from it"<<endl);
            // in case all our paths have been cut, apply the TRRs upwards
            sol += apply_trrs_upwards_after_cut(I, v);
          } else {
            // why is there no intrinsic invalid value of iterators?
            if(v->cyc_core_degree() == 2) { has_cyc_deg2 = true; last_cyc = v;}
            ++v;
          }
        } else ++v; // otherwise, just go on
      }
      // handle degerate inputs:
      if(!has_cyc_deg3){
        // if we are on a cycle, then apply PPRs from last_cyc
        if(has_cyc_deg2 && !allow_start_at_deg2){
          DEBUG2(cout << "okay, we're on a cycle! Time to allow using cyc-deg2 vertices"<<endl);
          // make sure the loop doesnt exit
          old_size = -1;
          // allow starting at cyc deg2
          allow_start_at_deg2 = true;
        } else {
          DEBUG2(cout << "okay, we're on a tree (or just failed to apply on a cycle)! Calling TRR6" << endl);
          trr6(I);
        }
      }

      DEBUG2(cout << "PRRs applied, solution so far: "<< sol << endl);
      // clean up and return the solution
    } while((old_size != sol.size()) && (!I.g.vertices.empty()));
    DEBUG2(cout << "returning from apply_prrs"<<endl);
    return sol;
  }
  
} // end of namespace


ostream& operator<<(ostream& os, const cr::path_info_t& t){
  if(!t.valid) cout << "INVALID ";
  os << "path: "<< *t.start << ".." << *t.end << " (len " << t.length << ") gen: " << t.generators << " sep: ";
  if(t.has_separator) cout << *t.separator; else cout << "-";
  return os << " pY: " << t.pendantYs;
}

