#include "trr.hpp"
#include "../util/graphs.hpp"

namespace cr{

  // apply Tree Reduction Rule X to v assuming all TRRs have been applied to the subtree below v

  // if nldeg(v) < 3   and   there is a ptwo, then delete the leaf of the ptwo
  // then, delete all but one leaf
  bool trr14_subtree(instance& I, const vertex_p& v){
    edgelist& leaves(v->trr_infos.leaves);
    edgelist& ptwos(v->trr_infos.ptwos);
    bool result = false;

    // first, if there is a P_2, then
    if((v->nldeg() < 3) && !ptwos.empty()){
      DEBUG2(cout << "applying TRR 4 to " << *v << " L: "<< leaves << " P2: "<<ptwos<< ")" <<endl);
      // delete its leaf - it's the unique leaf in (ptwos.begin())->partner's trr_infos :)
      I.g.delete_vertex(ptwos.front()->partner->trr_infos.leaves.front()->partner);
      // and register the middle vertex as new leaf
      leaves.push_back(*ptwos.begin());
      ptwos.pop_front();
      result = true;
    }

    // next, delete all but one leaf
    if(leaves.size() > 1){
      DEBUG2(cout << "TRR1: " << *v << " has " << leaves.size() << " leaves" << endl);
      // update the subtree_NH of v
      v->subtree_NH -= leaves.size()-1;

      // get iterator to second leaf and start deleting
      edgelist::iterator to_del = leaves.begin();
      while(++to_del != leaves.end())
        I.g.delete_vertex((*to_del)->partner);

      DEBUG2(cout << "cleared all leaves of " << *v << " except " << *leaves.begin() << endl);
      // update leaves, its prolly faster to just clear the list and reinsert the first element
      edge_p last_leaf(leaves.front());
      leaves.clear();
      leaves.push_back(last_leaf);
      result = true;
    }
    return result;
  }

  // while |Y|>0 and |Y|+|L|+|P|>1 delete some y of Y & reduce k by one
  solution_t trr2_subtree(instance& I, const vertex_p& v){
    uint leaves_and_ptwos = v->trr_infos.leaves.size() + v->trr_infos.ptwos.size();
    edgelist& ygraphs(v->trr_infos.ygraphs);
    list<vertex_p> to_del;
    solution_t sol;

    while(ygraphs.size()>0 && ygraphs.size() + leaves_and_ptwos > 1){
      DEBUG2(cout << "applying TRR2 to " << *v << " with " << v->subtree_NH << " of " << v->degree() << " neighbors in subtree" << endl);

      DEBUG3(cout << "TRR2: edge deletion: "<<*(ygraphs.front())<<endl);
      // delete the edge to the y-graph and delete the component containing the ygraph
      to_del.push_back(ygraphs.front()->partner);
      I.delete_edge(ygraphs.front(), sol);
      // keep subtree_NH of v up to date (we deleted an incident A-Bridge!)
      v->subtree_NH--;
      ygraphs.pop_front();
    }
    // delete the y-graphs that were cut off
    for(list<vertex_p>::iterator v = to_del.begin(); v != to_del.end(); ++v)
      I.g.delete_component(*v);

    return sol;
  }

  //  while |P|>2 delete a p in P and decrease k by one
  solution_t trr3_subtree(instance& I, const vertex_p& v){
    edgelist& ptwos(v->trr_infos.ptwos);
    list<vertex_p> to_del;
    solution_t sol;

    while(ptwos.size() > 2){
      DEBUG2(cout << "applying TRR3 to " << *v << "(" << ptwos.size() << " P2s)" <<endl);

      DEBUG3(cout << "TRR3: edge deletion: "<<*(ptwos.front())<<endl);
      // remove the connection edge and let delete_component do the rest
      to_del.push_back((*ptwos.begin())->partner);
      I.delete_edge(*(ptwos.begin()), sol);
      // keep subtree_NH of v up to date (we deleted an incident A-Bridge!)
      v->subtree_NH--;
      ptwos.pop_front();
    }
    // delete the edges that were cut off
    for(list<vertex_p>::iterator v = to_del.begin(); v != to_del.end(); ++v)
      I.g.delete_component(*v);
    
    return sol;
  }

  // if deg(w)=2 and |Y|=1 then delete w, delete y in Y and decrease k by one
  // note: we use the tclaws trr_info at the parent v of w to do this
  // note: please make sure trr2 has been applied to w before (that's what "_subtree" says...)
  solution_t trr5_subtree(instance& I, const vertex_p& v){
    if(!v->trr_infos.tclaws.empty()){
      DEBUG2(cout << "applying TRR5 to " << *v << "(subtree "<<v->subtree_NH << ", deg "<< v->degree() << ")" << endl);
      const edge_p& tc(v->trr_infos.tclaws.front());
      const vertex_p w(tc->partner);
      solution_t sol;

      // delete the edge {v,w} and throw away the component of w
      I.delete_edge(tc, sol);
      I.g.delete_component(w);
      
      // don't forget to update subtree_NH of v
      v->subtree_NH--;
      v->trr_infos.tclaws.pop_front();

      // note that some other TRR could now be applicable to v (for instance, TRR4)!
      sol += perform_trrs(I, v);
      return sol;
    }
    return solution_t();
  }

  // actualy do the application on a single vertex whose trr_infos are accurate!
  // return true if something happend and set local_operations to the number of edges deleted
  solution_t perform_trrs(instance& I, const vertex_p& v){
    // apply the tree reduction rules!
    // TODO: here, we ignore whether or not trr14 changed the graph since it has no impact outside of the TRRs (I hope)
    solution_t sol;
    // we remember the size of the solution before applying trrs, so we can perform as long as there are changes
    uint old_size;

    do{
      old_size = sol.size();
      trr14_subtree(I, v);
      sol += trr2_subtree(I, v);
      sol += trr3_subtree(I, v);
      sol += trr5_subtree(I, v);
    } while(old_size != sol.size());
    DEBUG2(cout << "done performing TRRs to " << *v << ", sol = " << sol << endl);
    return sol;
  }

  // update parents trr_infos from child (e := ( parent <--- child) )
  void update_trr_infos_from_child(const edge_p& e){
    const edge_p& to_child(e->get_reversed());
    const vertex_p& child(to_child->partner);
    const vertex_p& parent(e->partner);
    const trr_info_t& child_infos(child->trr_infos);

    child->set_parent(e);

    switch(child->degree()){
      case 1:
        parent->trr_infos.leaves.push_back(to_child);
        break;
      case 2:
        if(!child_infos.ygraphs.empty())
          parent->trr_infos.tclaws.push_back(to_child);
        else if(!child_infos.leaves.empty())
          parent->trr_infos.ptwos.push_back(to_child);
        break;
      default:
        if(child_infos.ptwos.size() > 1)
          parent->trr_infos.ygraphs.push_back(to_child);
        else // don't allow Ys with a leaf to register as P2 in the parent
          if(!child_infos.leaves.empty())
            parent->trr_infos.ptwos.push_back(to_child); // never use the same vertex for P2 and Y
    }
    DEBUG2(cout << "updated trr_infos of "<<*parent<<" with " << *child<< " ("<<child_infos<<") Now: "<<parent->trr_infos<<" (sub: "<<parent->subtree_NH<<" deg: "<<parent->degree()<<")"<<endl);
  }


 
  // return whether there is a vertex with nldeg>2 or a cycle in the component of v
  bool dfs_discover_nldegthree(const size_t dfs_id, const vertex_p& v, const vertex_p& parent){
    if(v->dfs_id != dfs_id){
      bool akku = false;
      v->dfs_id = dfs_id;
      // be sure to visit the whole connected component, even if a nldeg>2 vertex is found!
      // Otherwise, later searches in this component might miss this vertex
      // trr_infos cannot be trusted since a vertex with subtree_NH == degree() may not have some leaves registered
      if(v->nldeg() > 2){
        DEBUG2(cout << "found that "<<*v<<" has nldeg "<<v->nldeg()<< endl);
        akku = true;
      }
      for(edge_p e = v->adj_list.begin(); e != v->adj_list.end(); ++e)
        if(e->partner != parent)
          if(dfs_discover_nldegthree(dfs_id, e->partner, v))
            akku = true;// don't break because the cc must be marked
      return akku;
    } else return true;// discovered a cycle, which is equally as good
  }

  
  
  // Tree Reduction Rule 6 removes caterpillars
  bool trr6(instance& I){
    bool result = false;
    size_t dfs_id = I.g.get_dfs_id(); // we're slightly abusing dfs_ids to mark vertices
    list<vertex_p> to_del;

    for(vertex_p v = I.g.vertices.begin(); v != I.g.vertices.end(); ++v)
      if((v->degree() <= 1) && (v->dfs_id != dfs_id))
        if(!dfs_discover_nldegthree(dfs_id, v, v)){
          to_del.push_back(v);
          result = true;
        }

    for(list<vertex_p>::iterator v = to_del.begin(); v != to_del.end(); ++v)
      I.g.delete_component(*v);

    DEBUG2(cout << "done with TRR6 - " <<I.g.vertices.size() << " verts and "<<I.g.num_edges()<<" edges left"<< endl);
    return result;
  }



  // applies the tree reduction rules from v upwards after v got thrown out of the cyclic core
  // returns with v being the last found vertex (usually do_not_cross or a cyclic-core vertex)
  solution_t apply_trrs_upwards_after_cut(instance& I, vertex_p& v, const vertex_p do_not_cross){
    solution_t sol;

    DEBUG2(cout << "applying TRRs upwards from " << *v << " with: "<<v->trr_infos<<" (sub: "<<v->subtree_NH<<" deg: "<<v->degree()<<")"<<endl);
    DEBUG2(if(do_not_cross != I.g.vertices.end()) cout << "wont cross "<<do_not_cross<<endl);
    while(true){
      // apply the TRRs
      sol += perform_trrs(I, v);

      // if we meet do_not_cross, then return
      if(v == do_not_cross) return sol;

      // if we're on the cyclic core already, then stop processing
      if(v->is_on_cyclic_core()) return sol;
      // if we're at a new root, then stop
      if(v->degree() == v->subtree_NH) return sol;
   
      // my parent is my only neighbor who does not have me as parent
      DEBUG2(cout << "looking for parent of "<<*v << endl);
      bool success = false;
      const edge_p to_parent(v->get_parent(&success));
      DEBUG2(if(success) cout << "found parent "<<*to_parent->partner<< endl; else cout << "no parent found"<<endl;);
      // if none of my neighbors was on the cyclic core, then we're all done
      if(!success) return sol;

      // update the subtree_NH and trr_infos of parent
      const vertex_p& parent(to_parent->partner);
      parent->subtree_NH++;
      update_trr_infos_from_child(to_parent);

      // and continue
      v = parent;
    }
  }
  solution_t apply_trrs_upwards_after_cut(instance& I, vertex_p& v){
    return apply_trrs_upwards_after_cut(I, v, I.g.vertices.end());
  }



  // apply all trrs to all vertices, return number of deletions done
  // TODO: this is much like graph::update_subtree_NH, try to collapse them into one "bottom-up applicator"
  solution_t apply_trrs(instance& I){
    list<vertex_p> subtree_done;
    uint dfs_id(I.g.get_dfs_id());
    solution_t sol;
   
    // make sure the subtree_NH is up to date, we need to know the parents
    I.g.update_subtree_NH();

    // first, clear all trr_infos and save all leaves, since their (empty) subtrees are done
    for(vertex_p v = I.g.vertices.begin(); v != I.g.vertices.end(); ++v){
      v->trr_infos.clear();
      if(v->degree() == 1) subtree_done.push_back(v);
    }

    // this is more of a leaves-to-root BFS then a dfs, ah well...
    // As long as there are unconsidered vertices who have their subtree done, update their trr_infos
    while(!subtree_done.empty()){
      vertex_p v(subtree_done.front());
      subtree_done.pop_front();

      DEBUG2(cout << "considering " << (string)*v << " for TRRs ("<< subtree_done.size() <<" left)"<<endl);

      // do the actual application of tree reduction rules
      sol += perform_trrs(I, v);

      // mark v as considered
      v->dfs_id = dfs_id;

      // if v is not on the cyclic core, then
      if(v->has_parent()){
        // update the parent's trr_infos
        update_trr_infos_from_child(v->get_parent());
        // if also all subtrees of v's parent are done, then mark the parent to be considered
        if(v->get_parent()->partner->trr_infos_up_to_date()){
          DEBUG2(cout << "adding " << *v->get_parent()->partner << " (trrs: "<<v->get_parent()->partner->trr_infos.count() <<", subs: "<<v->get_parent()->partner->subtree_NH << ") to be considered" << endl);
          subtree_done.push_back(v->get_parent()->partner);
        }
      }
    }
    // don't forget TRR6
    trr6(I);

    return sol;
  }

};
